import array
import os
import random

from string import ascii_lowercase
from string import ascii_uppercase
from string import digits
from string import printable
from string import punctuation

from populous.compat import ascii_letters

from .base import Generator


class Text(Generator):

    def get_arguments(self, min_length=0, max_length=None, chars='<a-Z><0-9> ',
                      **kwargs):
        super(Text, self).get_arguments(**kwargs)

        self.min_length = min_length or 0
        self.max_length = max_length or 10000
        self.chars = self.get_chars(chars)

    def get_chars(self, description):
        categories = {
            '<a-Z>': ascii_letters,
            '<a-z>': ascii_lowercase,
            '<A-Z>': ascii_uppercase,
            '<0-9>': digits,
            '<spaces>': ' \t',
            '<printable>': printable,
            '<punctuation>': punctuation,
            '<newline>': '\n',
        }
        for category, chars in categories.items():
            description = description.replace(category, chars)

        return description

    def generate(self):
        chars = self.chars
        nb_chars = len(chars)

        while True:
            # get a random length for the string
            length = random.randint(self.min_length, self.max_length)
            # get a random array of short integers of the same length than
            # the final string
            rand_shorts = array.array('H', os.urandom(length * 2))
            # for each short in the array, get an element from the list
            # of possible chars
            yield ''.join([chars[short % nb_chars] for short in rand_shorts])



import importlib
import logging

import click
import click_log

from .loader import load_blueprint
from .exceptions import ValidationError, YAMLError, BackendError

logger = logging.getLogger('populous')
click_log.basic_config(logger)


def get_blueprint(files, **kwargs):
    try:
        return load_blueprint(*files, **kwargs)
    except (YAMLError, ValidationError) as e:
        raise click.ClickException(str(e))
    except Exception as e:
        raise click.ClickException("Unexpected error during the blueprint "
                                   "loading: {}".format(e))


@click.group()
@click.version_option()
@click_log.simple_verbosity_option(logger)
def cli():
    pass


@cli.group()
def run():
    pass


def _generic_run(modulename, classname, files, **kwargs):
    try:
        try:
            module = importlib.import_module(
                'populous.backends.' + modulename,
                package='populous.backends'
            )
            backend_cls = getattr(module, classname)
        except (ImportError, AttributeError):
            raise click.ClickException("Backend not found.")

        backend = backend_cls(**kwargs)
        blueprint = get_blueprint(files, backend=backend)

        try:
            with backend.transaction():
                blueprint.generate()

                logger.info("Closing DB transaction...")

        finally:
            backend.close()

        logger.info("Have fun!")

    except BackendError as e:
        raise click.ClickException(str(e))


@run.command()
@click.option('--host', help="Database host address")
@click.option('--port', type=int, help="Database host port")
@click.option('--db', help="Database name")
@click.option('--user', help="Postgresql user name used to authenticate")
@click.option('--password', help="Postgresql password used to authenticate")
@click.argument('files', nargs=-1, required=True)
def postgres(host, port, db, user, password, files):
    return _generic_run('postgres', 'Postgres', files, host=host, port=port,
                        db=db, user=user, password=password)


@cli.command()
def generators():
    """
    List all the available generators.
    """
    from populous import generators

    base = generators.Generator

    for name in dir(generators):
        generator = getattr(generators, name)

        if isinstance(generator, type) and issubclass(generator, base):
            name = generator.__name__
            doc = (generator.__doc__ or '').strip()

            if doc:
                click.echo("{} - {}".format(name, doc))
            else:
                click.echo(name)



from .base import Generator


class Value(Generator):

    def get_arguments(self, value=None, **kwargs):
        super(Value, self).get_arguments(**kwargs)

        self.value = self.parse_vars(value)

    def generate(self):
        while True:
            yield self.evaluate(self.value)



from __future__ import absolute_import

import yaml

from populous.exceptions import GenerationError
from .base import Generator


class Yaml(Generator):

    def get_arguments(self, value=None, to_json=False, **kwargs):
        super(Yaml, self).get_arguments(**kwargs)

        self.value = self.parse_vars(value)
        self.to_json = to_json

    def generate(self):
        if self.to_json:
            return self._generate_to_json()

        return self._generate()

    def _generate(self):
        while True:
            try:
                yield yaml.load(self.evaluate(self.value))
            except yaml.YAMLError as e:
                raise GenerationError(
                    "Item '{}', field '{}': Invalid YAML: {}"
                    .format(self.item.name, self.field_name, e)
                )

    def _generate_to_json(self):
        convert = self.blueprint.backend.json_adapter
        for value in self._generate():
            yield convert(value)



from collections import deque, OrderedDict


class Buffer(object):

    def __init__(self, blueprint, maxlen=1000):
        self.blueprint = blueprint
        self.backend = blueprint.backend
        self.maxlen = maxlen
        self.buffers = OrderedDict()

    def add(self, obj):
        item = self.blueprint.items[type(obj).__name__]
        buffer = self.get_buffer(item)
        buffer.append(obj)

        if len(buffer) == self.maxlen:
            self.write(item, buffer)

    def flush(self):
        while any(self.buffers.values()):
            for item_name, buffer in self.buffers.items():
                self.write(self.blueprint.items[item_name], buffer)

    def write(self, item, buffer=None):
        buffer = buffer or self.get_buffer(item)
        if not buffer:
            return
        ids = self.backend.write(
            item, tuple(item.db_values(obj) for obj in buffer)
        )
        item.batch_written(self, buffer, ids)
        buffer.clear()

    def get_buffer(self, item):
        try:
            return self.buffers[item.name]
        except KeyError:
            buffer = deque(maxlen=self.maxlen)
            self.buffers[item.name] = buffer
            return buffer



import random
from datetime import date
from datetime import datetime
from time import mktime, gmtime

from dateutil.parser import parse as dateutil_parse
from dateutil.tz import tzlocal

from .base import Generator


def to_timestamp(dt):
    if getattr(dt, 'tzinfo', None) is not None:
        dt = dt.astimezone(tzlocal())
    return mktime(dt.timetuple())


def parse_datetime(candidate):
    if isinstance(candidate, (datetime, date)):
        return candidate
    if isinstance(candidate, int):
        # we just assume this is a year
        candidate = str(candidate)
        # force the date to start at the 1st of January otherwise
        # dateutil takes the current day & month
        candidate = '01/01/' + candidate
    return dateutil_parse(candidate)


class DateTime(Generator):

    epoch_year = gmtime(0).tm_year

    def get_arguments(self, past=True, future=False, after=None, before=None,
                      **kwargs):
        super(DateTime, self).get_arguments(**kwargs)

        self.past = past
        self.future = future
        self.after = self.parse_vars(after)
        self.before = self.parse_vars(before)

    def generate(self):
        if not self.past:
            start = to_timestamp(datetime.now())
        else:
            # try not to go too far in the past
            start = to_timestamp(datetime(self.epoch_year, 1, 1))

        if self.future:
            # try not to go too far in the future
            stop = to_timestamp(datetime(2100, 1, 1))
        else:
            stop = to_timestamp(datetime.now())

        if self.before:
            stop = to_timestamp(parse_datetime(self.evaluate(self.before)))

        if self.after:
            start = to_timestamp(parse_datetime(self.evaluate(self.after)))

        while True:
            yield datetime.fromtimestamp(
                random.randint(start, stop)
            )


class Date(DateTime):

    def generate(self):
        for dt in super(Date, self).generate():
            yield dt.date()



from .base import BaseGenerator, Generator
from .choices import Choices
from .text import Text
from .date import DateTime, Date
from .value import Value
from .boolean import Boolean
from .integer import Integer
from .uuid import UUID
from .select import Select
from .email import Email
from .name import Name, FirstName, LastName
from .ip import IP
from .url import URL
from .store import Store
from .yaml import Yaml


__all__ = ['BaseGenerator', 'Generator', 'Choices', 'DateTime', 'Date',
           'Text', 'Value', 'Boolean', 'Integer', 'UUID', 'Select', 'Email',
           'Name', 'FirstName', 'LastName', 'IP', 'URL', 'Store', 'Yaml']



from django.conf.urls import include, url
from django.contrib import admin

urlpatterns = [
    # Examples:
    # url(r'^$', 'demo.views.home', name='home'),
    # url(r'^blog/', include('blog.urls')),

    url(r'^admin/', include(admin.site.urls)),
]



import random

from cached_property import cached_property
from faker import Factory

from populous.exceptions import GenerationError
from populous.exceptions import ValidationError
from populous.vars import Expression, parse_vars
from populous.bloom import BloomFilter

fake = Factory.create()


class BaseGenerator(object):

    def __init__(self, item, field_name, **kwargs):
        self.item = item
        self.field_name = field_name
        self.blueprint = item.blueprint

        # make a copy of the kwargs for inheritance
        self.kwargs = kwargs
        self.get_arguments(**kwargs)

    def __iter__(self):
        return self

    def next(self):
        return next(self.iterator)

    __next__ = next

    @cached_property
    def iterator(self):
        return iter(self.get_generator())

    def get_generator(self):
        return self.generate()

    def generate(self):
        raise NotImplementedError()

    def get_arguments(self, shadow=False, **kwargs):
        # should this field be written in the table?
        self.shadow = shadow

        if kwargs:
            raise ValidationError(
                "Item '{}', field '{}': Got extra param(s) for generator "
                "'{}': {}".format(
                    self.item.name, self.field_name, type(self).__name__,
                    ', '.join(kwargs.keys())
                )
            )

    def evaluate(self, value):
        if isinstance(value, Expression):
            return value.evaluate(**self.blueprint.vars)
        return value

    def parse_vars(self, value):
        return parse_vars(value)


class NullableMixin(object):

    def get_arguments(self, nullable=0, **kwargs):
        super(NullableMixin, self).get_arguments(**kwargs)

        if not nullable:
            self.nullable = 0
        elif nullable is True:
            self.nullable = 0.5
        else:
            self.nullable = self.parse_vars(nullable)

    def get_generator(self):
        if self.nullable:
            return self.generate_with_null()
        return super(NullableMixin, self).get_generator()

    def generate_with_null(self):
        generator = super(NullableMixin, self).get_generator()
        while True:
            if random.random() <= self.evaluate(self.nullable):
                yield None
            else:
                yield next(generator)


class UniquenessMixin(object):
    MAX_TRIES = 10000

    def get_arguments(self, unique=False, **kwargs):
        super(UniquenessMixin, self).get_arguments(**kwargs)

        self.unique = bool(unique)
        if isinstance(unique, (list, tuple)):
            self.unique_with = tuple(unique)
        elif unique and unique is not True:
            self.unique_with = (unique,)
        else:
            self.unique_with = None

        # default bloomfilter, used if for some reason a pre-filled
        # one is not assigned in Item.preprocess
        self.seen = BloomFilter()

    def get_generator(self):
        if self.unique:
            return self.generate_uniquely()
        return super(UniquenessMixin, self).get_generator()

    def generate_uniquely(self):
        seen = self.seen
        unique_with = self.unique_with
        tries = 0
        for value in super(UniquenessMixin, self).get_generator():
            if isinstance(value, tuple) and hasattr(value, 'id'):
                key = value.id
            else:
                key = value
            if unique_with:
                this = self.blueprint.vars['this']
                key = (key,) + tuple(getattr(this, f) for f in unique_with)
            if key in seen:
                tries += 1
                if tries > self.MAX_TRIES:
                    raise GenerationError(
                        "Item '{}', field '{}': Could not generate a "
                        "new unique value in {} tries. Aborting."
                        .format(self.item.name, self.field_name,
                                self.MAX_TRIES)
                    )
                continue
            tries = 0
            seen.add(key, check=False)
            yield value


class Generator(NullableMixin, UniquenessMixin, BaseGenerator):
    pass



import contextlib


class Backend(object):

    def __init__(self, *args, **kwargs):
        self.closed = False

    @property
    def json_adapter(self):
        import json
        return json.dumps

    @contextlib.contextmanager
    def transaction(self):
        yield

    def write(self, item, objs):
        pass

    def select_random(self, table, fields=None, where=None, max_rows=None):
        raise NotImplementedError()

    def select(self, table, fields):
        raise NotImplementedError()

    def close(self):
        self.closed = True

    def __del__(self):
        if not self.closed:
            try:
                self.close()
            except Exception:
                pass



from __future__ import absolute_import

import uuid

from .base import Generator


class UUID(Generator):

    def get_arguments(self, to_string=False, **kwargs):
        super(UUID, self).get_arguments(**kwargs)

        self.to_string = to_string

    def generate(self):
        if self.to_string:
            return (str(u) for u in self._generate())

        return self._generate()

    def _generate(self):
        while True:
            yield uuid.uuid4()



from .compat import bloomfilter


class BloomFilter(object):

    def __init__(self, capacity=1000, error_rate=0.000001):
        self._capacity = capacity
        self._error_rate = error_rate
        self._filters = []
        self._counter = 0
        self._filters.append(self._create_filter())

    def add(self, value, check=True):
        if check and value in self:
            return

        self._counter += 1
        if self._counter >= self._capacity:
            self._filters.append(self._create_filter())
            self._counter = 1

        f = self._filters[-1]
        f.add(value)

    def __contains__(self, value):
        for f in reversed(self._filters):
            if value in f:
                return True
        return False

    def _create_filter(self):
        if self._filters:
            self._capacity *= 4
            self._error_rate *= 0.9
        return bloomfilter.BloomFilter(self._capacity, self._error_rate)



import random

import jinja2


jinja_env = jinja2.Environment()
jinja_env.undefined = jinja2.StrictUndefined


# Fix an issue where 'random' filter is being inlined, resulting in the
# same value being generated every time.
# cf https://github.com/pallets/jinja/pull/478

@jinja2.contextfilter
def do_random(context, seq):
    """Return a random item from the sequence."""
    try:
        return random.choice(seq)
    except IndexError:
        return context.environment.undefined(
            'No random item, sequence was empty.'
        )


jinja_env.filters['random'] = do_random



import re
from operator import attrgetter

import jinja2

from populous.compat import basestring
from populous.exceptions import GenerationError
from populous.exceptions import ValidationError
from populous.jinja import jinja_env


VAR_REGEX = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*$')


def parse_vars(value):
    if not value or not isinstance(value, basestring):
        return value

    if value[0] == '$':
        value = value.strip()

        # find out if it is a '$var` or `$(expr)`
        if len(value) > 1 and value[1] == '(':
            if value[-1] != ')':
                raise ValidationError(
                    "Error parsing '{}': Missing ')'".format(value)
                )
            return JinjaValueExpression(value[2:-1])

        if not VAR_REGEX.match(value[1:]):
            raise ValidationError(
                "Error parsing '{}': This is not a valid value expression. "
                "You should either escape the $ with a '\\', or use the "
                "template syntax ('... {{{{ var }}}} ...')."
                .format(value)
            )
        return ValueExpression(value[1:])
    elif value.startswith('\$'):
        # the $ was escaped, remove the escape char
        value = value[1:]

    if '{{' in value or '{%' in value:
        # this might be a template expression
        # even if it's not, using a TemplateExpression will return
        # the same result and be almost as fast
        return TemplateExpression(value)

    return value


class Expression(object):

    def evaluate(self, **vars):
        raise NotImplementedError()


class ValueExpression(Expression):

    def __init__(self, value):
        self.value = value
        self.var, _, attrs = value.partition('.')
        if attrs:
            self.attrgetter = attrgetter(attrs)
        else:
            self.attrgetter = None
        self.attrs = attrs

    def evaluate(self, **vars_):
        try:
            var = vars_[self.var]
        except KeyError:
            raise GenerationError(
                "Error generating value '${}': '{}' is undefined"
                .format(self.value, self.var)
            )
        if self.attrgetter:
            try:
                return self.attrgetter(var)
            except AttributeError as e:
                # this message is not very informative ('type' object has no
                # attribute 'attr'), but we cannot do much better without
                # inspecting the calling frame
                raise GenerationError(
                    "Error generating value '${}': {}"
                    .format(self.value, e)
                )
        else:
            return var


class JinjaValueExpression(Expression):

    def __init__(self, value):
        self.value = value
        try:
            self.expression = jinja_env.compile_expression(
                value, undefined_to_none=False
            )
        except jinja2.TemplateError as e:
            raise ValidationError(
                "Error parsing '$({})': Invalid Jinja2 expression ({})"
                .format(value, e)
            )

    def evaluate(self, **vars_):
        try:
            value = self.expression(vars_)
            if isinstance(value, jinja2.Undefined):
                # trigger a UndefinedError
                str(value)
            return value
        except jinja2.UndefinedError as e:
            raise GenerationError(
                "Error generating value '$({})': {}".format(self.value, e)
            )


class TemplateExpression(Expression):

    def __init__(self, value=""):
        self.value = value
        try:
            self.template = jinja_env.from_string(value)
        except jinja2.TemplateError as e:
            raise ValidationError(
                "Error parsing template '{}': {}"
                .format(value, e)
            )

    def evaluate(self, **vars_):
        try:
            return self.template.render(vars_)
        except jinja2.UndefinedError as e:
            raise GenerationError(
                "Error generating template '{}': {}".format(self.value, e)
            )



from django.db import models


ABODE_TYPES = (
    ('SH', 'Small house'),
    ('H', 'House'),
    ('SB', 'Small building'),
    ('B', 'Building')
)

GENDERS = (
    ('M', 'Male'),
    ('F', 'Female')
)


class World(models.Model):
    name = models.CharField(max_length=30, unique=True)
    description = models.TextField(blank=True, default='')


class City(models.Model):
    world = models.ForeignKey(World, related_name='cities')

    name = models.CharField(max_length=40)


class Abode(models.Model):
    city = models.ForeignKey(City, related_name='abodes')
    owner = models.OneToOneField('Citizen', related_name='owned_abode',
                                 null=True)

    type = models.CharField(max_length=2, choices=ABODE_TYPES)


class Citizen(models.Model):
    abode = models.ForeignKey(Abode, related_name='citizens')

    first_name = models.CharField(max_length=20)
    middle_name = models.CharField(max_length=20, blank=True, default='')
    last_name = models.CharField(max_length=20)
    birth = models.DateField()
    gender = models.CharField(max_length=1, choices=GENDERS)


class Hobby(models.Model):
    name = models.CharField(max_length=40)
    citizens = models.ManyToManyField(Citizen)



from __future__ import absolute_import

import argparse
from collections import OrderedDict
from functools import partial

from django.apps import apps
from django.core import validators
from django.core.management.base import BaseCommand
from django.db import models

import yaml
from cached_property import cached_property


class SkipField(Exception):
    pass


class Command(BaseCommand):
    help = """
    This command will print a blueprint for Populous corresponding to all
    the models in the given apps.

    Example:
        $ django-admin populous auth myapp > blueprints/00-schema.yml

    The goal is not to generate a read-to-use blueprint, but to bootstrap
    it by doing all the heavy-lifting.
    In most cases the created blueprint will not be advanced
    enough to generate data respecting all your business or even technical
    constraints.

    Django fields are mapped to the most suitable generator, but
    sometimes, this might not be possible.
    In this case the field will be reprensented as a null value, and you'll
    have to override it. Please refer to Populous documentation to see how
    to do that.

    The fields not handled by this command are relational fields (ForeignKey,
    OneToOneField, ManyToManyField) and some specialized fields like
    CommaSeparatedIntegerField or FileField.

    This command identify unique fields, but not fields that are unique
    together. This is by design, because the command cannot know which
    is the field that should change when a duplicate is found during
    the generation.
    """

    def add_arguments(self, parser):
        # don't mess with the help text
        parser.formatter_class = argparse.RawDescriptionHelpFormatter
        parser.add_argument('apps', nargs='+')

    def handle(self, *args, **options):
        blueprint = {
            'items': list(self.export_models(options['apps']))
        }
        self.stdout.write(self.dump(blueprint))

    def dump(self, blueprint):
        class Dumper(yaml.SafeDumper):
            pass

        def _ordered_representer(dumper, data):
            return dumper.represent_mapping(
                yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
                data.items(),
                flow_style=False  # avoid dumping using inline representation
            )

        Dumper.add_representer(OrderedDict, _ordered_representer)
        return yaml.dump(blueprint, Dumper=Dumper, indent=2)

    def export_models(self, app_labels):
        # TODO: handle through table
        for app_label in app_labels:
            app = apps.get_app_config(app_label)
            for model in app.get_models():
                if model._meta.abstract:
                    continue
                yield self.export_model(model)

    def export_model(self, model):
        return OrderedDict((
            ('name', model.__name__),
            ('table', model._meta.db_table),
            ('fields', OrderedDict(self.export_fields(model))),
        ))

    def export_fields(self, model):
        for field in model._meta.get_fields():
            if not field.concrete or field.primary_key:
                continue
            try:
                yield self.export_field(field)
            except SkipField:
                continue

    def export_field(self, field):
        params = None

        if field.choices:
            params = OrderedDict((
                ('generator', 'Choices'),
                ('choices', [choice[0] for choice in field.choices])
            ))
        else:
            try:
                handler = next(
                    handler for field_type, handler in self.fields_mapping
                    if isinstance(field, field_type)
                )
            except StopIteration:
                raise SkipField()

            if handler:
                params = handler(field)
            else:
                params = None

        if params is not None:
            if 'nullable' not in params and field.null:
                params['nullable'] = True

            if 'unique' not in params and field.unique:
                params['unique'] = True

        # TODO: handle unique_for_date/month/year when filters are implemented

        return field.db_column or field.name, params

    @cached_property
    def fields_mapping(self):
        params = [
            (models.AutoField, self.skip),
            (models.BigIntegerField, self.handle_integer),
            (models.BinaryField, None),
            (models.BooleanField, self.handle_boolean),
            (models.CommaSeparatedIntegerField, None),
            (models.DateField, partial(self.handle_datetime, date=True)),
            (models.DateTimeField, self.handle_datetime),
            (models.DecimalField, None),
            (models.DurationField, None),
            (models.EmailField, partial(self.handle_text, generator='Email')),
            (models.FileField, None),
            (models.FilePathField, None),
            (models.FloatField, None),
            (models.ImageField, None),
            (models.IntegerField, self.handle_integer),
            (models.GenericIPAddressField, self.handle_ip),
            (models.NullBooleanField, self.handle_boolean),
            (models.PositiveIntegerField, self.handle_integer),
            (models.PositiveSmallIntegerField, self.handle_integer),
            (models.SlugField, None),
            (models.SmallIntegerField, self.handle_integer),
            (models.TextField, self.handle_text),
            (models.TimeField, None),
            (models.URLField, partial(self.handle_text, generator='URL')),
            (models.UUIDField, self.handle_uuid),
            (models.CharField, self.handle_text),
            (models.ForeignKey, None),
            (models.OneToOneField, None),
            (models.ManyToManyField, self.skip)
        ]

        try:
            # Django < 1.9
            params.append((models.IPAddressField, self.handle_ip))
        except AttributeError:
            pass

        try:
            # Django >= 1.10
            params.append((models.BigAutoField, self.skip))
        except AttributeError:
            pass

        return params

    def skip(self, field):
        raise SkipField()

    def handle_integer(self, field):
        params = OrderedDict((
            ('generator', 'Integer'),
        ))
        for validator in field.validators:
            if isinstance(validator, validators.MinValueValidator):
                params['min'] = validator.limit_value
            elif isinstance(validator, validators.MaxValueValidator):
                params['max'] = validator.limit_value
        return params

    def handle_boolean(self, field):
        return OrderedDict((
            ('generator', 'Boolean'),
        ))

    def handle_text(self, field, generator='Text'):
        params = OrderedDict((
            ('generator', generator),
            ('min_length', 0 if field.blank else 1),
        ))
        if field.max_length is not None:
            params['max_length'] = field.max_length
        return params

    def handle_datetime(self, field, date=False):
        params = OrderedDict((
            ('generator', 'Date' if date else 'DateTime'),
        ))
        if field.auto_now or field.auto_now_add:
            # make sure we don't set a created_at or updated_at field
            # to a future value
            # (this might be useless as it is the default of populous)
            params['past'] = True
            params['future'] = False
        return params

    def handle_ip(self, field):
        protocol = getattr(field, 'protocol', 'IPv4')
        return OrderedDict((
            ('generator', 'IP'),
            ('ipv4', protocol in ('both', 'IPv4')),
            ('ipv6', protocol in ('both', 'IPv6')),
        ))

    def handle_uuid(self, field):
        return OrderedDict((
            ('generator', 'UUID'),
            # This is a hack to avoid checking the value for uniqueness for
            # uuid fields, as it would be worthless
            ('unique', False),
        ))



import random

from .base import Generator


class Boolean(Generator):

    def get_arguments(self, ratio=0.5, **kwargs):
        super(Boolean, self).get_arguments(**kwargs)

        self.ratio = ratio

    def generate(self):
        while True:
            yield random.random() <= self.ratio



import os
import random
import contextlib
import uuid

try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache

from populous.compat import range
from populous.exceptions import BackendError
from .base import Backend

try:
    import psycopg2
    import psycopg2.extras
except ImportError:
    raise BackendError("You must install 'psycopg2' in order to use the "
                       "Postgresql backend")


class Postgres(Backend):
    def __init__(self, *args, **kwargs):
        super(Postgres, self).__init__(*args, **kwargs)

        self._current_cursor = None

        dbname = kwargs.pop('db', None) or os.environ.get('PGDATABASE')
        try:
            self.conn = psycopg2.connect(dbname=dbname, **kwargs)
        except psycopg2.DatabaseError as e:
            raise BackendError("Error connecting to Postgresql DB: {}"
                               .format(e))

        # authorize uuids objects in queries
        psycopg2.extras.register_uuid()
        # authorize dicts objects for hstore in queries
        try:
            psycopg2.extras.register_hstore(self.conn)
        except psycopg2.ProgrammingError:
            # this probably means that the hstore extension is not
            # installed in the db, no need to register it then
            pass

    @property
    def json_adapter(self):
        from psycopg2.extras import Json
        return Json

    @property
    @contextlib.contextmanager
    def cursor(self):
        if self._current_cursor:
            yield self._current_cursor
        else:
            with self.conn.cursor() as cursor:
                yield cursor

    @contextlib.contextmanager
    def transaction(self):
        with self.conn:
            with self.conn.cursor() as cursor:
                self._current_cursor = cursor
                yield cursor
                self._current_cursor = None

    def write(self, item, objs):
        with self.cursor as cursor:
            stmt = "INSERT INTO {} ({}) VALUES {} RETURNING {}".format(
                item.table,
                ", ".join(item.db_fields),
                ", ".join("({})".format(
                    ", ".join("%s" for _ in range(len(item.db_fields)))
                ) for _ in range(len(objs))),
                self.get_pk_column(item.table)
            )

            try:
                cursor.execute(stmt, tuple(v for vs in objs for v in vs))
            except psycopg2.DatabaseError as e:
                raise BackendError("Error during the generation of "
                                   "'{}': {}".format(item.name, e))

            return tuple(e[0] for e in cursor.fetchall())

    @lru_cache()
    def count(self, table, where=None):
        with self.cursor as cursor:
            cursor.execute("SELECT count(*) FROM {table} {where}".format(
                table=table,
                where="WHERE ({})".format(where) if where else '',
            ))
            return cursor.fetchone()[0]

    def select_random(self, table, fields=None, where=None, max_rows=1000):
        with self.cursor as cursor:
            count = self.count(table, where=where)

            cursor.execute(
                "SELECT {fields} FROM {table} "
                "WHERE random() < {proportion} {extra_where} {limit}"
                .format(
                    table=table,
                    fields=', '.join(fields),
                    proportion=min(max_rows / float(count), 1.0),
                    extra_where="AND ({})".format(where) if where else '',
                    limit="LIMIT {}".format(max_rows)
                )
            )

            results = cursor.fetchall()
            random.shuffle(results)
            return results

    def select(self, table, fields):
        with self.conn.cursor(str(uuid.uuid4())) as cursor:
            cursor.execute(
                "SELECT {fields} FROM {table}"
                .format(
                    fields=', '.join(fields),
                    table=table
                )
            )

            for value in cursor:
                yield value

    @lru_cache()
    def get_pk_column(self, table):
        with self.cursor as cursor:
            cursor.execute(
                """
SELECT a.attname FROM pg_index i
JOIN pg_attribute a ON a.attrelid = i.indrelid AND a.attnum = ANY(i.indkey)
WHERE i.indrelid = %s::regclass AND i.indisprimary;
                """,
                (table,)
            )
            return cursor.fetchone()[0]

    def close(self):
        if not self.closed:
            try:
                self.conn.close()
            finally:
                self.closed = True



from .base import Generator


class Select(Generator):

    def get_arguments(self, table=None, where=None, pk='id', **kwargs):
        super(Select, self).get_arguments(**kwargs)

        self.table = table
        self.where = self.parse_vars(where)
        self.pk = pk

    def generate(self):
        backend = self.blueprint.backend

        while True:
            where = self.evaluate(self.where)
            values = backend.select_random(self.table, fields=(self.pk,),
                                           where=where, max_rows=10000)
            for value in values:
                if self.evaluate(self.where) != where:
                    break
                yield value



import yaml

from .blueprint import Blueprint
from .exceptions import ValidationError
from .exceptions import YAMLError


BLUEPRINT_KEYS = ('items', 'vars')


def load_blueprint(*filenames, **kwargs):
    blueprint = Blueprint(**kwargs)

    for filename in filenames:
        content = _get_yaml(filename)
        try:
            _load_content(blueprint, content)
        except ValidationError as e:
            e.filename = filename
            raise e

    return blueprint


def _get_yaml(filename):
    with open(filename) as f:
        try:
            return yaml.load(f)
        except yaml.YAMLError as e:
            raise YAMLError(filename, str(e))


def _load_content(blueprint, content):
    if not content:
        return

    if not isinstance(content, dict):
        raise ValidationError(
            "The blueprint must be a dict containing the keys 'vars' and "
            "'items'. Got a '{}'.".format(type(content).__name__)
        )

    if any(key not in BLUEPRINT_KEYS for key in content.keys()):
        extra = set(content.keys()) - set(BLUEPRINT_KEYS)
        raise ValidationError(
            "Unknown key(s) in blueprint: '{}'. Possible keys are: '{}'."
            .format(', '.join(sorted(extra)), ', '.join(BLUEPRINT_KEYS))
        )

    _load_vars(blueprint, content.get('vars', {}))
    _load_items(blueprint, content.get('items', []))


def _load_vars(blueprint, vars_):
    if not isinstance(vars_, dict):
        raise ValidationError(
            "Blueprint vars must be a dict, not a {}."
            .format(type(vars_).__name__)
        )

    for name, value in vars_.items():
        blueprint.add_var(name, value)


def _load_items(blueprint, items):
    if not isinstance(items, list):
        raise ValidationError(
            "Blueprint items must be in a list, not a {}."
            .format(type(items).__name__)
        )

    for item in items:
        blueprint.add_item(item)



import random

from populous.exceptions import GenerationError

from .base import Generator


class Choices(Generator):

    def get_arguments(self, choices=(), **kwargs):
        super(Choices, self).get_arguments(**kwargs)

        if isinstance(choices, str):
            self.from_var = True
            self.choices = self.parse_vars(choices)
        else:
            self.from_var = False
            self.choices = choices

    def generate(self):
        if self.from_var:
            generator = self._generate_from_var
        else:
            if not self.choices:
                raise GenerationError(
                    "The choices for field '{}' of item '{}' are empty."
                    .format(self.field_name, self.item.name)
                )
            generator = self._generate_from_list

        while True:
            yield generator()

    def _generate_from_list(self):
        return self.evaluate(random.choice(self.choices))

    def _generate_from_var(self):
        try:
            return random.choice(self.evaluate(self.choices))
        except IndexError:
            if self.nullable:
                return None
            raise GenerationError(
                "The choices for field '{}' of item '{}' are empty, and "
                "the field is not nullable."
                .format(self.field_name, self.item.name)
            )



class PopulousError(Exception):
    pass


class YAMLError(PopulousError):

    def __init__(self, filename, problem):
        super(YAMLError, self).__init__(
            "Error parsing '{}': {}".format(filename, problem)
        )


class ValidationError(PopulousError):

    def __init__(self, *args, **kwargs):
        super(ValidationError, self).__init__(*args, **kwargs)
        self.filename = kwargs.pop('filename', None)

    def __str__(self):
        msg = super(ValidationError, self).__str__()
        if self.filename:
            return "File '{}': {}".format(self.filename, msg)
        else:
            return msg


class GenerationError(PopulousError):
    pass


class BackendError(PopulousError):
    pass





class ItemFactory(object):

    def __init__(self, item, parent=None):
        self.item = item
        self.blueprint = item.blueprint
        self._generated = {}

        if parent:
            self._generated[item.count.by] = parent

    def __getattribute__(self, name):
        try:
            return super(ItemFactory, self).__getattribute__(name)
        except AttributeError:
            generated = self._generated

            if name in generated:
                return generated[name]
            elif name in self.item.fields:
                value = self._get_value(name)
                generated[name] = value
                return value
            else:
                raise

    def generate(self):
        obj = self.item.namedtuple._make(
            getattr(self, name) for name in self.item.namedtuple._fields
        )
        self.clear()
        return obj

    def clear(self):
        parent = self.item.count.by
        if parent:
            self._generated = {parent: self._generated[parent]}
        else:
            self._generated = {}

    def _get_value(self, name):
        return next(self.item.fields[name])



from populous.exceptions import ValidationError

from .base import Generator, fake


class Name(Generator):

    def get_arguments(self, gender=None, max_length=None, **kwargs):
        super(Name, self).get_arguments(**kwargs)

        self.gender = self.parse_vars(gender)
        self.max_length = max_length

    def _get_provider(self):
        gender = self.evaluate(self.gender)

        if gender not in ('M', 'F', None):
            raise ValidationError(
                "Gender must be either 'M', 'F' or null. Got '{}'"
                .format(gender)
            )

        if gender == 'F':
            return fake.name_female
        elif gender == 'M':
            return fake.name_male
        else:
            return fake.name

    def generate(self):
        provider = self._get_provider()
        while True:
            value = provider()
            if self.max_length and len(value) > self.max_length:
                continue
            yield value


class FirstName(Name):

    def _get_provider(self):
        if self.gender == 'F':
            return fake.first_name_female
        elif self.gender == 'M':
            return fake.first_name_male
        else:
            return fake.first_name


class LastName(Generator):

    def get_arguments(self, max_length=None, **kwargs):
        super(LastName, self).get_arguments(**kwargs)

        self.max_length = max_length

    def generate(self):
        while True:
            name = fake.last_name()
            if self.max_length and len(name) > self.max_length:
                continue
            yield name



import random

from .base import Generator, fake


class IP(Generator):

    def get_arguments(self, ipv4=True, ipv6=True, **kwargs):
        super(IP, self).get_arguments(**kwargs)
        self.ipv4 = ipv4
        self.ipv6 = ipv6

    def generate(self):
        if self.ipv4 and self.ipv6:
            return self.generate_both()
        elif self.ipv4:
            return self.generate_ipv4()
        else:
            return self.generate_ipv6()

    def generate_ipv4(self):
        while True:
            yield fake.ipv4()

    def generate_ipv6(self):
        while True:
            yield fake.ipv6()

    def generate_both(self):
        while True:
            if random.random() >= 0.5:
                yield fake.ipv4()
            else:
                yield fake.ipv6()



from .base import Generator


class Store(Generator):

    def get_arguments(self, **kwargs):
        kwargs['shadow'] = True
        super(Store, self).get_arguments(**kwargs)

    def generate(self):
        while True:
            yield []



import copy
import logging
import random
from collections import OrderedDict
from collections import namedtuple

from cached_property import cached_property

from populous import generators
from populous.compat import zip
from populous.compat import range
from populous.exceptions import ValidationError
from populous.factory import ItemFactory
from populous.vars import parse_vars
from populous.vars import Expression
from populous.vars import ValueExpression

logger = logging.getLogger('populous')

ITEM_KEYS = ('name', 'parent', 'table', 'count', 'fields', 'store_in')
COUNT_KEYS = ('number', 'by', 'min', 'max')


class Item(object):

    def __init__(self, blueprint, name, table, parent=None, store_in=None):
        self.blueprint = blueprint

        if parent:
            table = table or parent.table
            store_in = store_in or parent._store_in

        if not name:
            # if any of our parent had a name we would have inherited it
            raise ValidationError("Items without a parent must have a name.")

        if not table:
            raise ValidationError(
                "Item '{}' does not have a table.".format(name)
            )

        self.name = name
        self.table = table
        self.count = Count(
            number=0, by=None, min=None, max=None, blueprint=blueprint
        )
        self.fields = OrderedDict()
        self._store_in = store_in
        self._set_store_in(store_in)

        self.add_field('id', 'Value', value=None, shadow=True)

        if parent:
            for name, field in parent.fields.items():
                self.add_field(name, type(field).__name__, **field.kwargs)
            self.count = parent.count

        if parent:
            self.ancestors = parent.ancestors
            # only add parent to ancestors if it will not be generated
            # otherwise we would generate our children for both our parent
            # and us
            if parent.count.number == 0:
                self.ancestors += [parent.name]
        else:
            self.ancestors = []

    @cached_property
    def namedtuple(self):
        fields = tuple(self.fields.keys())
        if self.count.by:
            fields += (self.count.by,)
        return namedtuple(self.name, fields)

    @cached_property
    def db_fields(self):
        return tuple(
            name for name, field in self.fields.items() if not field.shadow
        )

    def _set_store_in(self, store_in):
        if not store_in:
            self.store_in_global = {}
            self.store_in_item = {}
            return

        self.store_in_global = {
            name: parse_vars(expression)
            for name, expression in store_in.items()
            if not name.startswith('this.')
        }
        # create the global var in the blueprint
        for name in self.store_in_global:
            self.blueprint.vars.setdefault(name, [])

        self.store_in_item = {
            ValueExpression(name): parse_vars(expression)
            for name, expression in store_in.items()
            if name.startswith('this.')
        }
        # create a "store" generator on the targeted item
        for expr in self.store_in_item:
            target_name, store_name = expr.attrs.rsplit('.')[-2:]
            try:
                target = self.blueprint.items[target_name]
            except KeyError:
                raise ValidationError(
                    "Error in 'store_in' section in item '{}': "
                    "The item '{}' does not exist."
                    .format(self.name, target_name)
                )
            target.add_field(store_name, 'Store')

    def add_field(self, name, generator, **params):
        if not generator:
            parent_field = self.fields.get(name, None)
            if parent_field:
                # if the field already exists and a new generator
                # has not been set, we use the same generator and the
                # same default params than the existing one
                generator = type(parent_field).__name__
                parent_kwargs = copy.deepcopy(parent_field.kwargs)
                parent_kwargs.update(params)
                params = parent_kwargs
            else:
                raise ValidationError(
                    "Field '{}' in item '{}' must either be a value, or "
                    "a dict with a 'generator' key."
                    .format(name, self.name)
                )

        try:
            generator_cls = getattr(generators, generator)
        except AttributeError:
            raise ValidationError(
                "Item '{}', field '{}': Generator '{}' does not exist."
                .format(self.name, name, generator)
            )
        self.fields[name] = generator_cls(self, name, **params)

    def add_count(self, number=None, by=None, min=None, max=None):
        number = parse_vars(number)
        min = parse_vars(min)
        max = parse_vars(max)

        for key, value in zip(('number', 'min', 'max'), (number, min, max)):
            if value is None:
                continue
            if isinstance(value, int):
                if value < 0:
                    raise ValidationError(
                        "Item '{}' count: {} must be positive."
                        .format(self.name, key)
                    )
            elif not isinstance(value, Expression):
                raise ValidationError(
                    "Item '{}' count: {} must be an integer or a variable "
                    "(got: '{}')."
                    .format(self.name, key, type(value).__name__)
                )

        current_count = self.count
        if current_count:
            # this item already has a count, merge it
            by = by or current_count.by
            if min is None and max is None:
                number = number or current_count.number
            if number is None:
                if min is None:
                    min = current_count.min
                if max is None:
                    max = current_count.max

        if min is not None or max is not None:
            if number is not None:
                raise ValidationError(
                    "Item '{}' count: Cannot set 'number' and 'min/max'."
                    .format(self.name)
                )

        # treat None as 0
        min = min or 0
        max = max or 0

        if isinstance(min, int) and isinstance(max, int) and min > max:
            raise ValidationError(
                "Item '{}' count: Min is greater than max."
                .format(self.name)
            )

        self.count = Count(
            number=number, by=by, min=min, max=max, blueprint=self.blueprint
        )

    def preprocess(self):
        seen_fields = self.blueprint.seen[self.table]
        db_fields = []
        to_fill = []
        for field in self.fields.values():
            if field.shadow or not getattr(field, 'unique'):
                continue

            index = len(db_fields)
            if field.unique_with:
                fields = (field.field_name,) + tuple(field.unique_with)
                if fields not in seen_fields:
                    db_fields += fields
                    to_fill.append((field, slice(index, index + len(fields))))
                field.seen = seen_fields[fields]
            else:
                field_name = field.field_name
                if field_name not in seen_fields:
                    db_fields.append(field_name)
                    to_fill.append((field, index))
                field.seen = seen_fields[field_name]

        if not db_fields:
            return

        for values in self.blueprint.backend.select(self.table, db_fields):
            for field, index in to_fill:
                field.seen.add(values[index], check=False)

    def batch_written(self, buffer, batch, ids):
        logger.info("{:>5} {} written".format(len(batch), self.name))

        objs = tuple(e._replace(id=id) for (e, id) in zip(batch, ids))
        self.store_values(objs)
        self.generate_dependencies(buffer, objs)

    def store_values(self, objs):
        def _get_values(expression):
            for obj in objs:
                self.blueprint.vars['this'] = obj
                yield expression.evaluate(**self.blueprint.vars)
            del self.blueprint.vars['this']

        for name, expression in self.store_in_global.items():
            store = self.blueprint.vars[name]
            store.extend(_get_values(expression))

        if self.store_in_item:
            for obj in objs:
                self.blueprint.vars['this'] = obj
                for name_expr, value_expr in self.store_in_item.items():
                    store = name_expr.evaluate(**self.blueprint.vars)
                    store.append(value_expr.evaluate(**self.blueprint.vars))
            del self.blueprint.vars['this']

    def generate(self, buffer, count, parent=None):
        factory = ItemFactory(self, parent=parent)

        for i in range(count):
            self.blueprint.vars['this'] = factory
            obj = factory.generate()
            del self.blueprint.vars['this']

            buffer.add(obj)

    def generate_dependencies(self, buffer, batch):
        # generate items having a "count by" on this item
        # or one of its ancestors
        names = frozenset(self.ancestors) | {self.name}
        for item in self.blueprint.items.values():
            by = item.count.by
            if by in names:
                for obj in batch:
                    self.blueprint.vars[by] = obj
                    count = item.count()
                    del self.blueprint.vars[by]
                    if count:
                        item.generate(buffer, count, parent=obj)
                buffer.write(item)

    def db_values(self, obj):
        return tuple(getattr(obj, field) for field in self.db_fields)


class Count(namedtuple('Count', COUNT_KEYS + ('blueprint',))):
    __slots__ = ()

    def evaluate(self, value):
        if isinstance(value, Expression):
            return value.evaluate(**self.blueprint.vars)
        return value

    def __call__(self):
        if self.number is not None:
            return self.evaluate(self.number)
        return random.randint(self.evaluate(self.min), self.evaluate(self.max))



import logging

from collections import OrderedDict, defaultdict

from populous.bloom import BloomFilter
from populous.buffer import Buffer
from populous.compat import basestring
from populous.exceptions import ValidationError
from populous.item import Item, COUNT_KEYS, ITEM_KEYS


logger = logging.getLogger('populous')


class Blueprint(object):

    def __init__(self, items=None, vars_=None, backend=None):
        self.items = OrderedDict(items or {})
        self.vars = vars_ or {}
        self.backend = backend

        # a dict containing {<table>: {<fields>: bloom filter}}
        # to store the existing values of unique fields
        self.seen = defaultdict(lambda: defaultdict(BloomFilter))

    def add_var(self, name, value):
        self.vars[name] = value

    def add_item(self, description):
        if not isinstance(description, dict):
            raise ValidationError(
                "A blueprint item must be a dict, not a '{}'"
                .format(type(description).__name__)
            )

        if any(key not in ITEM_KEYS for key in description.keys()):
            extra = set(description.keys()) - set(ITEM_KEYS)
            raise ValidationError(
                "Unknown key(s) '{}'. Possible keys are '{}'."
                .format(', '.join(extra), ', '.join(ITEM_KEYS))
            )

        name = description.pop('name', None)
        parent = None
        if name and name in self.items:
            # this item already exists, this mean we are trying
            # to override it
            if description.get('parent', name) != name:
                raise ValidationError(
                    "Re-defining item '{}' while setting '{}' as parent "
                    "is ambiguous.".format(name, description['parent'])
                )
            description['parent'] = name

        if description.get('parent'):
            parent_name = description['parent']
            if parent_name not in self.items:
                raise ValidationError(
                    "Parent '{}' does not exist.".format(parent_name)
                )
            parent = self.items[parent_name]
            if not name:
                name = parent.name

        store_in = description.get('store_in')
        if store_in:
            if not isinstance(store_in, dict):
                raise ValidationError(
                    "'store_in' must be a dict, not a '{}'"
                    .format(type(store_in).__name__)
                )

        item = Item(
            blueprint=self, name=name, table=description.get('table'),
            parent=parent, store_in=store_in
        )

        fields = description.get('fields', {})
        if not isinstance(fields, dict):
            raise ValidationError(
                "Fields must be a dict, not a '{}'"
                .format(type(fields).__name__)
            )
        for field_name, attrs in fields.items():
            if isinstance(attrs, dict):
                generator = attrs.pop('generator', None)
                params = attrs
            else:
                # If we didn't get a dict, this is a fixed value
                generator = 'Value'
                params = {'value': attrs}

            item.add_field(field_name, generator, **params)

        count = description.get('count')
        if count is None:
            pass
        elif isinstance(count, int):
            item.add_count(number=count)
        elif isinstance(count, basestring) and count[0] == '$':
            item.add_count(number=count)
        else:
            if not isinstance(count, dict):
                raise ValidationError(
                    "The count of item '{}' must be an integer or a dict."
                    .format(name)
                )

            if any(key not in COUNT_KEYS for key in count.keys()):
                extra = set(count.keys()) - set(COUNT_KEYS)
                raise ValidationError(
                    "Unknown key(s) '{}' in count of item '{}'. Possible "
                    "keys are '{}'."
                    .format(', '.join(extra), name, ', '.join(COUNT_KEYS))
                )

            item.add_count(**count)

        self.items[name] = item

    def preprocess(self):
        for item in self.items.values():
            item.preprocess()

    def generate(self):
        logger.info("Getting existing unique values...")

        self.preprocess()

        buffer = Buffer(self)

        logger.info("Starting generation...")

        for item in self.items.values():
            if item.count.by:
                # we only create the items with no dependency,
                # the others will be created on the fly
                continue

            item.generate(buffer, item.count())

        # write everything left in the buffer
        buffer.flush()

        logger.info("Generation done.")



import sys

from setuptools import setup, find_packages

import populous


def readme():
    with open('README.rst') as f:
        return f.read()


requirements = [
    "click",
    "click-log>=0.2.0",
    "cached-property",
    "dateutils",
    "Faker",
    "future",
    "Jinja2",
    "PyYAML",
]

if sys.version_info < (3, 2):
    requirements.append('functools32')

if sys.version_info < (3,):
    requirements.append('peloton_bloomfilters')
else:
    requirements += [
        'cython',
        'pybloomfiltermmap3',
    ]


setup(
    name="populous",
    version=populous.__version__,
    url=populous.__url__,
    description="Populate your database with god-like powers",
    author=populous.__author__,
    author_email=populous.__author_email__,
    license=populous.__license__,
    long_description=readme(),
    packages=find_packages(exclude=['demo']),
    install_requires=requirements,
    extras_require={
        'tests': ['tox', 'pytest', 'pytest-mock', 'flake8'],
    },
    entry_points={
        'console_scripts': [
            'populous = populous.__main__:cli'
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Utilities",
    ],
    keywords='populous populate database',
    zip_safe=False,
)



import random

from populous.vars import parse_vars
from .base import Generator


class Integer(Generator):

    def get_arguments(self, min=0, max=(2 ** 32) - 1, to_string=False,
                      **kwargs):
        super(Integer, self).get_arguments(**kwargs)

        self.min = parse_vars(min)
        self.max = parse_vars(max)
        self.to_string = to_string

    def generate(self):
        if self.to_string:
            return (str(e) for e in self._generate())
        return self._generate()

    def _generate(self):
        while True:
            yield random.randint(
                self.evaluate(self.min),
                self.evaluate(self.max)
            )



import sys

from builtins import range
from builtins import zip
from six import string_types as basestring


PY2 = sys.version_info[0] == 2


if PY2:
    import peloton_bloomfilters as bloomfilter
else:
    import pybloomfilter as bloomfilter


if PY2:
    from string import letters as ascii_letters
else:
    from string import ascii_letters


__all__ = [
    'ascii_letters',
    'basestring',
    'bloomfilter',
    'PY2',
    'range',
    'zip',
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-03-01 12:59
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Abode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[(b'SH', b'Small house'), (b'H', b'House'), (b'SB', b'Small building'), (b'B', b'Building')], max_length=2)),
            ],
        ),
        migrations.CreateModel(
            name='Citizen',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('first_name', models.CharField(max_length=20)),
                ('middle_name', models.CharField(blank=True, default=b'', max_length=20)),
                ('last_name', models.CharField(max_length=20)),
                ('birth', models.DateField()),
                ('gender', models.CharField(choices=[(b'M', b'Male'), (b'F', b'Female')], max_length=1)),
                ('abode', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='citizens', to='demo.Abode')),
            ],
        ),
        migrations.CreateModel(
            name='City',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=40)),
            ],
        ),
        migrations.CreateModel(
            name='Hobby',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=40)),
                ('citizens', models.ManyToManyField(to='demo.Citizen')),
            ],
        ),
        migrations.CreateModel(
            name='World',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, unique=True)),
                ('description', models.TextField(blank=True, default=b'')),
            ],
        ),
        migrations.AddField(
            model_name='city',
            name='world',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cities', to='demo.World'),
        ),
        migrations.AddField(
            model_name='abode',
            name='city',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='abodes', to='demo.City'),
        ),
        migrations.AddField(
            model_name='abode',
            name='owner',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='owned_abode', to='demo.Citizen'),
        ),
    ]



"""
Populate your database with god-like powers
"""

__author__ = "PeopleDoc"
__author_email__ = "yannick.peroux@people-doc.com"
__copyright__ = "Copyright 2017, PeopleDoc"
__license__ = "MIT"
__url__ = 'http://github.com/novafloss/populous'
__version__ = '0.2.2'
